//
//  VDB Update Script for Windows
//
//  Copyright (c) 2004-2012 VirusBuster Ltd. All rights reserved.
//


/* Class VDB */

/** Creates a new object that represents a VDB.
  *
  * A VDB can be stored in a local folder or to be fetched from a remote site.
  * It can be fully functional or only partially downloaded. The latter is the
  * case when updating an existing VDB. Partial VDB needs to be installed over
  * a full one to become usable. */
function VDB (path)
{
    this.localPath = null;
    this.remoteURL = null;

    /* Check if VDB is on a remote site. If so, make XML locally available */
    if (   path.indexOf ("ftp://")  == 0
        || path.indexOf ("http://") == 0)
    {
        this.remoteURL = new String (path);

        /* Create temporary download directory */
        var tmpDir = FSO.BuildPath (FSO.GetSpecialFolder (2),
                                    FSO.GetTempName ());
        FSO.CreateFolder (tmpDir);

        /* Get XML */
        download (path + "/vdb.xml", FSO.BuildPath (tmpDir, "vdb.xml"));

        this.localPath = new String (tmpDir);
    }
    else
        this.localPath = new String (path);
    
    if (!FSO.FileExists (FSO.BuildPath (this.localPath, "vdb.xml")))    
        throw new Error (1, "VDB not found at " + this.localPath);
}

/** Deletes local VDB folder */
VDB.prototype.trash = function ()
{
    FSO.DeleteFolder (this.localPath);
    this.localPath = null;
}

/** Checks if local VDB files are valid according to the XML. Optionally,
  * missing files can be excluded from validation (partial mode). */
VDB.prototype.validate = function (isPartial)
{
    var cmd = FSO.BuildPath (".", "vdbxdiff.exe") + " -v "
            + (isPartial ? "--omit=missing " : "")
            + FSO.BuildPath (this.localPath, "vdb.xml");

    var err = exec (cmd);
    if (err != 0)
        throw new Error (1, "Invalid VDB set");
}

/** Downloads extra VDB files compared to an other VDB. */
VDB.prototype.download = function (diff)
{
    var e = new Enumerator (diff.added);

    for (e.moveFirst (); !e.atEnd (); e.moveNext ())
    {
        print ("  " + e.item ());
        download (this.remoteURL + "/" + e.item (),
                  FSO.BuildPath (this.localPath, e.item ()));
    }
}

/** Creates a local copy of the VDB. */
VDB.prototype.copyTo = function (dest)
{
    if (!FSO.FolderExists (dest))
        FSO.CreateFolder (dest);

    FSO.CopyFile (FSO.BuildPath (this.localPath, "*"), dest + "\\");

    return new VDB (dest);
}

/** Merges VDB into an other one. */
VDB.prototype.install = function (oldVDB, diff)
{
    var dest = oldVDB.localPath + "\\";

    /* Copy new files */
    var e = new Enumerator (diff.added);
    for (e.moveFirst (); !e.atEnd (); e.moveNext ())
    {
        var src = FSO.BuildPath (this.localPath, e.item ());
        FSO.CopyFile (src, dest);
    }

    /* Remove obsolete ones */
    e = new Enumerator (diff.deleted);
    for (e.moveFirst (); !e.atEnd (); e.moveNext ())
    {
        FSO.DeleteFile (FSO.BuildPath (dest, e.item ()));
    }

    /* Update XML */
    FSO.CopyFile (FSO.BuildPath (this.localPath, "vdb.xml"), dest);
}


/* Class XDiff */

/** Compares two VDBs and stores the differences in a new object. */
function XDiff (VDB1, VDB2)
{
    /* Difference classes */
    this.added   = new Array ();
    this.deleted = new Array ();
    this.invalid = new Array ();
    this.unknown = new Array ();

    /* Run VDB XML diff tool */
    var p1 = VDB1 != null ? FSO.BuildPath (VDB1.localPath, "vdb.xml") : "NUL";
    var p2 = FSO.BuildPath (VDB2.localPath, "vdb.xml");
    var cmd = FSO.BuildPath (".", "vdbxdiff.exe") + " --separator=| "
            + p1 + " " + p2;

    var err = exec (cmd, this.procDiff, this);
    if (err != 0)
    {
        throw new Error (1, "vdbxdiff failed between\n" + p1
                                            + "\nand\n" + p2);
    }
}

/** Processes a line of vdbxdiff output. */
XDiff.prototype.procDiff = function (line)
{
    var a = line.split ("|", 2);
    a[1] = FSO.GetFileName (a[1]);

    switch (a[0])
    {
        case '+':
            this.added.push (a[1]);
            break;

        case '-':
            this.deleted.push (a[1]);
            break;

        case '!':
            this.invalid.push (a[1]);
            break;

        case '?':
            this.unknown.push (a[1]);
            break;

        default:
            throw new Error (1, "Unexpected line in diff: " + line);
            break;
    }
}

XDiff.prototype.toString = function ()
{
    return "added:   " + this.added.join   ("\n") + "\n"
         + "deleted: " + this.deleted.join ("\n") + "\n"
         + "invalid: " + this.invalid.join ("\n") + "\n"
         + "unknown: " + this.unknown.join ("\n");
}


/* Helpers */

/** Downloads a remote file to a local folder. */
function download (from, to)
{
    var err = exec (FSO.BuildPath (".", "download.exe") + " " + from
                                                        + " " + to);
    if (err != 0)
    {
        throw new Error (1, "Unable to download\n"
                          + "from " + from + "\nto " + to);
    }
}

/** Outputs a message. Always prints when running on a console. GUI pop-ups
  * must be explicitly requested. */
function print (msg, isPopup)
{
    if (isOnConsole)
        WScript.Echo (msg);
        
    if (isPopup)
        Shell.Popup (msg, 60);
}

/** Runs a command and has its output processed.
  *
  * Output is consumed simultaneously to prevent blocking the external
  * process.
  *
  * Output lines are passed to the user-defined function 'procFunc' line by
  * line. Processing function is called with its current object changed to
  * 'procObj'. This is needed for calling methods properly. */
function exec (cmd, procFunc, procObj)
{
    /* Issue command */
    var exec = Shell.Exec (cmd);

    while (true)
    {
        /* Check if output is unavailable */
        if (exec.StdOut.atEndOfStream)
        {
            /* Check if the external process is running */
            if (exec.Status == 0)
            {
                /* Wait for output */
                WScript.Sleep (1000);
                continue;
            }
            else
                break;
        }

        /* Process output */
        var line = exec.StdOut.ReadLine ();

        /* Check if processing function supplied */
        if (procFunc != null)
        {
            /* Determine proper way to call */
            if (procObj != null)
                /* Method */
                procFunc.call (procObj, line);
            else
                /* Plain function */
                procFunc (line);
        }
    }

    return exec.ExitCode;
}


/* Main Logic */

/* Import external objects */
Shell  = new ActiveXObject ("WScript.Shell");
FSO    = new ActiveXObject ("Scripting.FileSystemObject");

/* Detect script engine */
var isOnConsole = (WScript.FullName.indexOf ("cscript.exe") != -1);

/* Force running on console to prevent annoying pop-ups */
if (!isOnConsole)
{
    Shell.Run (FSO.BuildPath (WScript.Path, "cscript.exe") + " "
             + WScript.ScriptFullName);
    WScript.Quit ();
}

var oldVDB = null;
var newVDB = null
var backup = null;
var exitMsg = new String();

/* Old VDB is not required */
try
    { oldVDB = new VDB ("vdb"); }
catch (e)
    { oldVDB = null; }

try
{
    newVDB = new VDB ("http://update.virusbuster.hu/pub12/vbuster/vdb12/");
    
    print ("Checking for updates");
    var diff = new XDiff (oldVDB, newVDB);
    
    /* Check if VDB is up to date */
    if (   diff.added.length   > 0
        || diff.deleted.length > 0)
    {
        /* Get new files */
        print ("Downloading " + diff.added.length + " file"
              + (diff.added.length > 1 ? "s": ""));
        newVDB.download (diff);
        newVDB.validate (true);

        /* Check if it's an update or a new installation */
        if (oldVDB != null)
        {
            /* Backup current VDB */
            print ("Backing up");
            backup = oldVDB.copyTo (FSO.BuildPath (oldVDB.localPath, "backup"));

            /* Update */
            print ("Installing");
            newVDB.install (oldVDB, diff);
            oldVDB.validate (false);
        }
        else
            newVDB.copyTo ("vdb");

        exitMsg = "Successfully updated VDB";
    }
    else
        exitMsg = "Latest VDB has already been installed";
}
catch (e)
{
    /* Restore backup */
    if (backup != null)
    {
        print ("Restoring backup");
        backup.copyTo (oldVDB.localPath);
    }

    exitMsg = "Failed to update VDB\n\nError: " + e.message;
}

/* Cleanup */
print ("Cleaning up");
if (newVDB != null)
    newVDB.trash ();

print (exitMsg, true);

/* End of Main Logic */
